#!/usr/bin/perl
package AlignmentReasoning;  #the package name
use base "Exporter";
our @EXPORT = qw(labelChunkAlignments, labelChunkAlignments_MachineLearning, JRip_Rules);  #telling what we should export

use WordNet::QueryData 1.40;
use WordQuery;
use List::Util qw( min max );

use strict;
use warnings;

###############################
#---------- Global Variables ----------
###############################
my $wq;

##################################
#---------- Constants ----------
##################################
#SPE test results
my $A_More_Specific_Than_B = 1;
my $B_More_Specific_Than_A = 2;

#POS Contants - based on 
#   This is not a complete list of the possible POS tags but rather a short list for how they will be used for alignment reasoning
my $POS_Noun = "((NN)|(NNP)|(NNPS)|(NNS))";  #Non-pronoun nouns
my $POS_Verb = "((VB)|(VBD)|(VBG)|(VBN)|(VBP)|(VBZ))";  #verbs (all tenses and forms)
my $POS_Adjective = "((JJ)|(JJR)|(JJS)|(CD))"; #adjectives and cardinal numbers
my $POS_Adverb = "((RB)|(RBR)|(RBS)|(RP))";  #adverbs
my $POS_Conjunction = "((CC)|(IN))";  #conjunctions
my $POS_ProperNoun = "((NNP)|(NNPS))";  #proper noun


#          ---------- Alignment Reasoning Entry Point ----------
#Assigns alignment reasoning to a set of two pre-aligned and chunked sentances
#      Input   - \@words1 = the chunked sentance1 (an array of chunk references)
#              - \@words2 = the chunked sentance2 (an array of chunk references)
#              - \@chunks1 = the chunk information for the longer of the two sentences
#              - \@chunks2 = the chunk information for the shorter of the two sentences
#              - \@posTags1 = the posTags for sentance1 (an array of posTag arrays)
#              - \@posTags2 = the posTags for sentance2 (an array of posTag arrays)
#              - \@alignments = the chunk to chunk alignments for sentance1 and sentance2 (an array of chunk indeces)
#              - $wq = a WordQuery Instance (saves time so that it doesn't have to get initialized at each call)
#              - \@chunkSimilarities  = the scores for each chunk pair (an array of numbers)
#      Returns - \@alignmentReasons, \@scores
#                 - \@alignmentReasons = labels for each alignment (per SEMEVAL specifications)
#                 - \@alignmentScores = similairtiy scores for each alignment (per SEMEVAL specifications)
#     Notes:  A "chunk" is an array of tokens
sub labelChunkAlignments_rules
{
    #grab the input and initialize
    my @words1 = @{ shift @_ };
    my @words2 = @ { shift @_ };
    my @chunks1 = @ { shift @_ };
    my @chunks2 = @ { shift @_ };
    my @posTags1 = @ { shift @_ };
    my @posTags2 = @ { shift @_ };
    my @alignments = @ { shift @_ };
    my @chunkSimilarities = @ { shift @_ };
    $wq = shift;

    #set up data structures
    my $numChunks1 = (max @chunks1) + 1;
    my @alignmentReasons = ();
    my @alignmentScores = ();

    #make all alignments of 0 (meaning no alignment)
    $alignmentReasons[0] = "NOALI";
    $alignmentScores[0] = "0";

    #begin the loop#loop through the sentance and assign alignment reasoning for each chunk
    #    Also modify the scores to meet the requirements
    for (my $i = 0; $i < $numChunks1; $i++) {

	#skip no alignments
	if ($alignments[$i] != 0)  {  #subtract 1 be
	    #create the chunks for sentence 1 and 2 that correspond to this alignment
	    my @result = &createChunks(\@words1, \@words2, 
				       \@posTags1, \@posTags2,
				       \@chunks1, \@chunks2,
				       \@alignments, \@chunkSimilarities, $i);
	    my @chunkA = @{ shift @result };
	    my @chunkB = @{ shift @result };
	    my @posTagsA = @{ shift @result };
	    my @posTagsB = @{ shift @result };
	    my $simScore = shift @result;

	    #get the alignment reason and updated score
	    my ($alignmentReason, $alignmentScore) = &labelChunkAlignment_Rules(\@chunkA, \@chunkB,
										\@posTagsA, \@posTagsB, 
										$simScore);
	    push @alignmentReasons, $alignmentReason;
	    push @alignmentScores, $alignmentScore;
	}
	else {
	    push @alignmentReasons, "NOALI";
	    push @alignmentScores, 0;
	}
    }

    #return the results
    return (\@alignmentReasons, \@alignmentScores);
}

#---------------------------------------------------------------------------

sub JRip_Rules
{
    #grab input
    my @words1 = @{ shift @_ };
    my @words2 = @{ shift @_ };
    my @chunks1 = @{ shift @_ };
    my @chunks2 = @{ shift @_ };
    my @posTags1 = @{ shift @_ };
    my @posTags2 = @{ shift @_ };
    my @alignments = @{ shift @_ };
    my @chunkSimilarities = @{ shift @_ };
    $wq = shift;
    my $ruleSetIndex = shift;

    #assign an alignment reason and score for each alignment pair
    my @alignmentReasons = ();
    push @alignmentReasons, 'NOALI';
    my $numAlignments = (max @chunks1) + 1;
    for (my $i = 0; $i < $numAlignments; $i++) {

	#only chunks that have an alignment need to be converted
	my $index = $i+1;
	if ($alignments[$i] > 0) {
	    #create the chunk data structures
	    my @result = &createChunks(\@words1, \@words2, 
				       \@posTags1, \@posTags2, 
				       \@chunks1, \@chunks2,
				       \@alignments, \@chunkSimilarities, $i);
	    my @chunkA = @{ shift @result };
	    my @chunkB = @{ shift @result };
	    my @posTagsA = @{ shift @result };
	    my @posTagsB = @{ shift @result };
	    my $simScore = shift @result;

	    #chunkA may have size of 0 if there are multiple that chunk has multiple
	    #   alignments (the value is overwritten).  ChunkB is added to be safe
	    if (scalar @chunkA > 0 && scalar @chunkB > 0) {
		#add each pair to the training set (features will be extracted there)
		my @attributes = &getAttributes(\@chunkA, \@chunkB, \@posTagsA, \@posTagsB, $simScore);

		#determine the alignment reason
	        my $reason = &labelAlignment_JRip_all(\@attributes);
		push @alignmentReasons, $reason;

	    }
	    else {
		push(@alignmentReasons, 'NOALI');
	    }
	}
	else {
	    push (@alignmentReasons, 'NOALI');
	}
    }

    #return the created attributes
    return @alignmentReasons;
}

#Applies rules created from WEKA's JRIP
sub labelAlignment_JRip_all {
    my @features = @{ shift @_ };

    #OPPO Rules
    if(($features[15] >= 1) && ($features[50] >= 1)) { 
	return 'OPPO';
    }
    if(($features[15] >= 1) && ($features[8] >= 1) 
       && ($features[12] >= 1) && ($features[33] >= 1)) { 
	return 'OPPO';
    }
    if(($features[49] >= 0.5) && ($features[14] >= 1) 
       && ($features[28] <= 1)) { 
	return 'OPPO';
    }
    if(($features[15] >= 1) && ($features[52] >= 0.5) 
       && ($features[1] >= 0.5)) { 
	return 'OPPO';
    }

    #SPE2 Rules
    if(($features[2] >= 0.333333) && ($features[31] <= -1) 
       && ($features[18] >= 1) && ($features[37] <= 0) 
       && ($features[29] >= 0.571429)) { 
	return 'SPE2';
    }

    #SPE1 Rules
    if(($features[31] >= 1) && ($features[18] >= 1) 
       && ($features[2] <= 0) && ($features[13] >= 1) 
       && ($features[27] <= 4) && ($features[68] <= 0)) { 
	return 'SPE1';
    }
    if(($features[31] >= 1) && ($features[34] >= 2) 
       && ($features[29] >= 0.714286)) { 
	return 'SPE1';
    }
    if(($features[32] >= 1) && ($features[35] >= 0.888889) 
       && ($features[2] <= 0) && ($features[7] >= 1)) { 
	return 'SPE1';
    }
    if(($features[1] >= 0.25) && ($features[31] >= 1) 
       && ($features[12] <= 0) && ($features[1] <= 0.666667) 
       && ($features[48] <= 0.2) && ($features[35] <= 1) 
       && ($features[29] >= 0.666667)) { 
	return 'SPE1';
    }

    #SIMI Rules
    if(($features[2] >= 0.5) && ($features[1] >= 0.333333) 
       && ($features[20] >= 1)) { 
	return 'SIMI';
    }

    #return the defaul, EQUI
    return 'EQUI';
}

#---------------------------------------------

#converts a chunk pair to a feature vector for use in external machine learning
sub convertToFeatures
{
    #grab input
    my @words1 = @{ shift @_ };
    my @words2 = @{ shift @_ };
    my @chunks1 = @{ shift @_ };
    my @chunks2 = @{ shift @_ };
    my @posTags1 = @{ shift @_ };
    my @posTags2 = @{ shift @_ };
    my @alignments = @{ shift @_ };
    my @alignmentScores = @{ shift @_ };
    my @alignmentReasons = @{ shift @_ };
    my @chunkSimilarities = @{ shift @_ };
    $wq = shift;

    #assign an alignment reason and score for each alignment pair
    my @allAttributes = ();
    my $numAlignments = (max @chunks1) + 1;
    for (my $i = 0; $i < $numAlignments; $i++) {

	#only chunks that have an alignment need to be converted
	my $index = $i+1;
	if ($alignmentReasons[$index] ne 'NOALI') {

	    #create the chunk data structures
	    my @result = &createChunks(\@words1, \@words2, 
				       \@posTags1, \@posTags2, 
				       \@chunks1, \@chunks2,
				       \@alignments, \@chunkSimilarities, $i);
	    my @chunkA = @{ shift @result };
	    my @chunkB = @{ shift @result };
	    my @posTagsA = @{ shift @result };
	    my @posTagsB = @{ shift @result };
	    my $simScore = shift @result;

	    #chunkA may have size of 0 if there are multiple that chunk has multiple
	    #   alignments (the value is overwritten).  ChunkB is added to be safe
	    if (scalar @chunkA > 0 && scalar @chunkB > 0) {
		#add each pair to the training set (features will be extracted there)
		my @attributes = &getAttributes(\@chunkA, \@chunkB, \@posTagsA, \@posTagsB, $simScore);
		push @allAttributes, \@attributes;
	    }
	}
    }

    #return the created attributes
    return \@allAttributes;
}


#create chunk and posTags based on the original words and the alignment scores
#create the chunks for sentence 1 and 2 that correspond to this alignment
sub createChunks
{
    #grab the input
    my @words1 = @{ shift @_ };
    my @words2 = @{ shift @_ };
    my @posTags1 = @{shift @_ };
    my @posTags2 = @{ shift @_ };
    my @chunks1 = @{ shift @_ };
    my @chunks2 = @{ shift @_ };
    my @alignments = @{ shift @_ };
    my @chunkSimilarities = @{ shift @_ };
    my $chunk1Number = shift;

    #create chunk and posTags for words1
    my @chunkA = ();
    my @posTagsA = ();
    for (my $i = 0; $i < scalar @words1; $i++) {
	if ($chunks1[$i] == $chunk1Number) {
	    push @chunkA, $words1[$i];
	    push @posTagsA, $posTags1[$i];
	}
    }

    #create chunk and posTags for words2
    my @chunkB = ();
    my @posTagsB = ();
    for (my $i = 0; $i < scalar @words2; $i++) {
	if ($chunks2[$i] == ($alignments[$chunk1Number] -1)) {
	    push @chunkB, $words2[$i];
	    push @posTagsB, $posTags2[$i];
	}
    }

    #return the chunks
    return ( \@chunkA, \@chunkB, \@posTagsA, \@posTagsB, $chunkSimilarities[$chunk1Number] );
}


#Computes the attributes for a feature vector
sub getAttributes
{
    #grab variables
    my @chunkA = @{ shift @_ };
    my @chunkB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my $chunkSimilarity = shift;

    #get content words
    my ($aRef, $aPOSRef) = &getContentWords(\@chunkA, \@posTagsA);
    my ($bRef, $bPOSRef) = &getContentWords(\@chunkB, \@posTagsB);
    my @contentWordsA = @{ $aRef };
    my @contentWordsB = @{ $bRef };
    my @contentPOSTagsA = @{ $aPOSRef };
    my @contentPOSTagsB = @{ $bPOSRef };
    my $numContentWordsA = scalar @contentWordsA;
    my $numContentWordsB = scalar @contentWordsB;

    #get unmatched content words
    ($aRef, $bRef, $aPOSRef, $bPOSRef) = &getUnmatchedContentWords(\@contentWordsA, \@contentWordsB, 
								   \@contentPOSTagsA, \@contentPOSTagsB);
    my @unmatchedContentWordsA = @{ $aRef; };
    my @unmatchedContentWordsB = @{ $bRef };
    my @unmatchedContentPOSTagsA = @{ $aPOSRef };
    my @unmatchedContentPOSTagsB = @{ $bPOSRef };
    my $numUnmatchedContentWordsA = scalar @unmatchedContentWordsA;
    my $numUnmatchedContentWordsB = scalar @unmatchedContentWordsB;

    #grab attributes
    ( my $aHasNounContentWord, my $aHasVerbContentWord, my $aHasAdjectiveContentWord, my $aHasAdverbContentWord ) 
	=  &determineContentTypes(\@contentPOSTagsA);
    ( my $bHasNounContentWord, my $bHasVerbContentWord, my $bHasAdjectiveContentWord, my $bHasAdverbContentWord ) 
	= &determineContentTypes(\@contentPOSTagsB);
    ( my $haveSharedNoun, my $haveSharedVerb ) 
	= &shareWordTypes(\@contentWordsA, \@contentWordsB, \@contentPOSTagsA, \@contentPOSTagsB);
    ( my $aHasHypernymOfB, my $bHasHypernymOfA ) 
	= &containHypernyms(\@contentWordsA, \@contentWordsB, \@posTagsA, \@posTagsB);
    
    #compute matched percents
    my $unmatchedContentAPercent = 0;
    if ($numContentWordsA > 0) {
	$unmatchedContentAPercent = $numUnmatchedContentWordsA/$numContentWordsA,
    }
    my $unmatchedContentBPercent = 0;
    if ($numContentWordsB > 0) {
	$unmatchedContentBPercent = $numUnmatchedContentWordsB/$numContentWordsB,
    }

    #compute if the only content word is a noun
    my $onlyContentAIsNoun = 0;
    if ($numContentWordsA == 1) {
	$onlyContentAIsNoun = &isNoun($contentPOSTagsA[0]);
    }
    my $onlyContentBIsNoun = 0; 
    if ($numContentWordsB == 1) {
	$onlyContentBIsNoun = &isNoun($contentPOSTagsB[0]);
    }

    #compute if the only unmatched content word is a proper noun
    my $onlyUnmatchedContentAIsProperNoun = 0;
    if ($numUnmatchedContentWordsA == 1) {
	$onlyUnmatchedContentAIsProperNoun = &isProperNoun($unmatchedContentPOSTagsA[0]);
    }
    my $onlyUnmatchedContentBIsProperNoun = 0;
    if ($numUnmatchedContentWordsB == 1) {
	$onlyUnmatchedContentBIsProperNoun = &isProperNoun($unmatchedContentPOSTagsB[0]);
    }
   
    #set features
    my @attributes = ();
    push @attributes, $chunkSimilarity;              #0
    push @attributes, $unmatchedContentAPercent;     #1
    push @attributes, $unmatchedContentBPercent;     #2
    push @attributes, &hasConjunction(\@posTagsA);   #3
    push @attributes, &hasConjunction(\@posTagsB);   #4
    push @attributes, &hasNumericQuantity(\@chunkA); #5
    push @attributes, &hasNumericQuantity(\@chunkB); #6
    push @attributes, &hasLocation(\@chunkA);        #7
    push @attributes, &hasLocation(\@chunkB);        #8

    push @attributes, &hasDateTime(\@chunkA);        #9
    push @attributes, &hasDateTime(\@chunkB);        #10

    push @attributes, $aHasNounContentWord;          #11
    push @attributes, $aHasVerbContentWord;          #12
    push @attributes, $aHasAdjectiveContentWord;     #13
    push @attributes, $aHasAdverbContentWord;        #14

    push @attributes, &containAntonymContentWords(\@contentWordsA, \@contentWordsB, 
						  \@contentPOSTagsA, \@contentPOSTagsB);  #15

    push @attributes, $aHasHypernymOfB;             #16
    push @attributes, $bHasHypernymOfA;             #17
    push @attributes, $haveSharedNoun;              #18
    push @attributes, $haveSharedVerb;              #19
    push @attributes, &hasNumericQuantity(\@unmatchedContentWordsA) 
	&& &hasNumericQuantity(\@unmatchedContentWordsB);            #20

    push @attributes, $onlyUnmatchedContentAIsProperNoun;            #21
    push @attributes, $onlyUnmatchedContentBIsProperNoun;            #22
    push @attributes, $onlyContentAIsNoun;                           #23
    push @attributes, $onlyContentBIsNoun;                           #24




    #length features
    my $numContentA = scalar @contentWordsA;
    my $numContentB = scalar @contentWordsB;
    my $numWordsA = scalar @posTagsA;  #well this isnt the number of words because it also includes puntuation
    my $numWordsB = scalar @posTagsB;

    #raw counts
    push @attributes, $numContentA;               #25
    push @attributes, $numContentB;               #26
    push @attributes, $numWordsA;                 #27
    push @attributes, $numWordsB;                 #28

    #ratios/differences
    if ($numWordsA > 0) {
	push @attributes, $numContentA/$numWordsA;    #29
    }
    else {
	push @attributes, 0;
    }
    if ($numWordsB > 0) {
	 push @attributes, $numContentB/$numWordsB;   #30
    }
    else {
	push @attributes, 0;
    }
    push @attributes, $numContentA-$numContentB;  #31
    push @attributes, $numWordsA - $numWordsB;    #32
    push @attributes, abs $numContentA-$numContentB;   #33
    push @attributes, abs $numWordsA - $numWordsB;   #34
    if ($numWordsB > 0 && $numContentB > 0 && $numWordsA > 0) {
	push @attributes, ($numContentA/$numWordsA)
	    /($numContentB/$numWordsB);  #35
    }
    else {
	push @attributes, 0;
    }


    #word type features
    my ($numNounsA, $numVerbsA, $numAdjectivesA, 
	$numAdverbsA, $numConjunctionsA, $numProperNounsA) 
	= &countWordTypes(@posTagsA);
    my ($numNounsB, $numVerbsB, $numAdjectivesB, 
	$numAdverbsB, $numConjunctionsB, $numProperNounsB) 
	= &countWordTypes(@posTagsB);

    push @attributes, $numNounsA;  #36
    push @attributes, $numNounsB;  #37
    push @attributes, $numVerbsA;  #38
    push @attributes, $numVerbsB;  #39
    push @attributes, $numAdjectivesA;   #40
    push @attributes, $numAdjectivesB;   #41
    push @attributes, $numAdverbsA;    #42
    push @attributes, $numAdverbsB;    #43
    push @attributes, $numConjunctionsA;  #44
    push @attributes, $numConjunctionsB;  #45
    push @attributes, $numProperNounsA;   #46
    push @attributes, $numProperNounsB;   #47

    #ratios/differences
    if ($numContentA > 0) {
	push @attributes, $numNounsA/$numContentA;  #48
	push @attributes, $numVerbsA/$numContentA;   #49
	push @attributes, $numAdjectivesA/$numContentA;   #50
	push @attributes, $numAdverbsA/$numContentA;   #51
	push @attributes, $numConjunctionsA/$numContentA;  #52
	push @attributes, $numConjunctionsA/$numWordsA;   #52
	push @attributes, $numProperNounsA/$numContentA;  #53
    }
    else {
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
    }

    if ($numContentB > 0) {
	push @attributes, $numNounsB/$numContentB;  #54
	push @attributes, $numVerbsB/$numContentB;  #55
	push @attributes, $numAdjectivesB/$numContentB;   #56
	push @attributes, $numAdverbsB/$numContentB;   #57
	push @attributes, $numConjunctionsB/$numContentB;   #58
	push @attributes, $numConjunctionsB/$numWordsB;   #59
	push @attributes, $numProperNounsB/$numContentB;   #60
    }
    else {
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
	push @attributes, 0;
    }

    push @attributes, $numNounsA- $numNounsB;   #61
    push @attributes, $numVerbsA - $numVerbsB;   #62
    push @attributes, $numAdjectivesA - $numAdjectivesB;   #63
    push @attributes, $numAdverbsA - $numAdverbsB;  #64
    push @attributes, $numConjunctionsA - $numConjunctionsB;  #65
    push @attributes, $numProperNounsA - $numProperNounsB;  #66

    push @attributes, abs $numNounsA - $numNounsB;  #67
    push @attributes, abs $numVerbsA - $numVerbsB;   #68
    push @attributes, abs $numAdjectivesA - $numAdjectivesB;   #69
    push @attributes, abs $numAdverbsA - $numAdverbsB;  #70
    push @attributes, abs $numConjunctionsA - $numConjunctionsB;  #71
    push @attributes, abs $numProperNounsA - $numProperNounsB;  #72

    #return the list of attributes
    return @attributes;
}


#counts the types of each word based on the POS tags
sub countWordTypes {

    #grab input
    my @posTags = shift;

    #initialize counts
    my $numNouns = 0;
    my $numVerbs = 0;
    my $numAdjectives = 0;
    my $numAdverbs = 0;
    my $numConjunctions = 0;
    my $numProperNouns = 0;
    
    #cycle through pos tags and get the counts
    for (my $i = 0; $i < scalar @posTags; $i++) {
	if (&isNoun($posTags[$i])) {
	    $numNouns++;

	    if (&isProperNoun($posTags[$i])) {
		$numProperNouns++;
	    }
	}
	elsif (&isVerb($posTags[$i])) {
	    $numVerbs++;
	}
	elsif (&isAdjective($posTags[$i])) {
	    $numAdjectives++;
	}
	elsif(&isAdverb($posTags[$i])) {
	    $numAdverbs++;
	}
	elsif (&isConjunction($posTags[$i])) {
	    $numConjunctions++;
	}
    }

    return ($numNouns, $numVerbs, $numAdjectives, $numAdverbs, $numConjunctions, $numProperNouns);
}


#determines if the two chunks contain hypernyms of each other
sub containHypernyms
{
    #grab the input
    my @contentWordsA = @{ shift @_  };
    my @contentWordsB = @{ shift @_  };
    my @contentPOSTagsA = @{ shift @_  };
    my @contentPOSTagsB = @{ shift @_  };

    #find hypernyms (a to b)
    my $aHasHypernymOfB = 0;
    my $bHasHypernymOfA = 0;
    for(my $i = 0; $i < scalar @contentWordsA; $i++) {
	for (my $j = 0; $j < scalar @contentWordsB; $j++) {
	    if (&isHypernym($contentWordsA[$i], $contentWordsB[$j], $contentPOSTagsA[$i], $contentPOSTagsB[$j])) {
		$aHasHypernymOfB = 1;
	    }
	}
    }
    #find hypernyms (b to a)
    for(my $i = 0; $i < scalar @contentWordsB; $i++) {
	for (my $j = 0; $j < @contentWordsA; $j++) {
	    if (&isHypernym($contentWordsB[$i], $contentWordsA[$j], $contentPOSTagsB[$i], $contentPOSTagsA[$j])) {
		$bHasHypernymOfA = 1;
	    }
	}
    }

    #return the values
    return ( $aHasHypernymOfB, $bHasHypernymOfA );
}


#determines if the two chunks share a noun or a verb  
sub shareWordTypes
{
    #grab the input
    my @contentWordsA = @{ shift @_ };
    my @contentWordsB = @{ shift @_ };
    my @posTagsA = @{ shift @_  };
    my @posTagsB = @{  shift @_ };

    #compare each word
    my $haveSharedNoun = 0;
    my $haveSharedVerb = 0;
    for (my $i = 0; $i < scalar @contentWordsA; $i++) {
	for (my $j = 0; $j < scalar @contentWordsB; $j++) {
	    if (&wordsEquivalent($contentWordsA[$i], $contentWordsB[$j], $posTagsA[$i], $posTagsB[$j])) {
		if (&isNoun($posTagsA[$i])) { #we can assume the POS tags are the same since the words are equivalent
		    $haveSharedNoun = 1;
		} 
		elsif (&isVerb($posTagsA[$i])) {
		    $haveSharedVerb = 1;
		}
	    }
	}
    }
    
    #return the values
    return ( $haveSharedNoun, $haveSharedVerb );
}


#determines if a set of pos tags contains a noun, verb, adjective, and adverb
sub determineContentTypes
{
    #grab the input
    my @posTags =  @{ shift @_ };

    #initialize all to false
    my $hasNoun = 0;
    my $hasVerb = 0;
    my $hasAdjective = 0;
    my $hasAdverb = 0;

    #loop through all the pos tags and determine if each type is present
    foreach my $posTag(@posTags) {
	if (&isNoun($posTag)) {
	    $hasNoun = 1;
	}
	elsif (&isVerb($posTag)) {
	    $hasVerb = 1;
	} 
	elsif (&isAdjective($posTag)) {
	    $hasAdjective = 1;
	}
	elsif (&isAdverb($posTag)) {
	    $hasAdverb = 1;
	}
    }
    
    #retyurn the results
    return ( $hasNoun, $hasVerb, $hasAdjective, $hasAdverb );
}

#Assigns an alignment reason label and score to a pair of chunks using rules found in the NeroSim paper
#   Input - @chunkA, @chunkB, @posTagsA, @posTagsB, $similarityScore
#           = two tokenized chunks, their part of speech (POS) tags and the pre-computed similarity score
#   Returns - a chunk alignment reason (per SEMEVAL requirements) and an alignment score (per SEMEVAL requirements)
sub labelChunkAlignment_Rules
{
    #grab variables
    my @chunkA = @{ shift @_ };
    my @chunkB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my $chunkSimilarity = shift;

    #get the content words in A and B
    my ($aRef, $aPOSRef) = &getContentWords(\@chunkA, \@posTagsA);
    my ($bRef, $bPOSRef) = &getContentWords(\@chunkB, \@posTagsB);
    my @contentWordsA = @{ $aRef };
    my @contentWordsB = @{ $bRef };
    my @contentPOSTagsA = @{ $aPOSRef };
    my @contentPOSTagsB = @{ $bPOSRef };

    #get a set of non-matching content words (content words that are not similar in either sentance)
    ($aRef, $bRef, $aPOSRef, $bPOSRef) = &getUnmatchedContentWords(\@contentWordsA, \@contentWordsB, 
								   \@contentPOSTagsA, \@contentPOSTagsB);
    my @unmatchedContentWordsA = @{ $aRef; };
    my @unmatchedContentWordsB = @{ $bRef };
    my @unmatchedContentPOSTagsA = @{ $aPOSRef };
    my @unmatchedContentPOSTagsB = @{ $bPOSRef };

    #get conditions (used in multiple rules)
    my @conditions = &applyConditions(\@chunkA, \@chunkB, 
				      \@posTagsA, \@posTagsB, 
				      \@contentWordsA, \@contentWordsB, 
				      \@contentPOSTagsA, \@contentPOSTagsB);
    #check for equivalency
    if (&equivalentRule(scalar @contentWordsB, scalar @contentWordsA, 
			\@unmatchedContentPOSTagsA, \@unmatchedContentPOSTagsB, 
			\@conditions, $chunkSimilarity)) {
	return ("EQUI", 5);
    }

    #check for opposites
    if (&oppositeRule(@conditions)) {
	return ("OPPO", 0);
    }
 
    #check for one chunk more specific than the other
    my ($result, $alignmentScore) = &specificityRule(\@chunkA, \@chunkB, 
						     \@posTagsA, \@posTagsB, 
						     \@contentWordsA, \@contentWordsB, 
						     \@contentPOSTagsA, \@contentPOSTagsB);
    if($result == $A_More_Specific_Than_B) {
	return ("SPE1", $alignmentScore);
    } 
    elsif($result == $B_More_Specific_Than_A) {
	return ("SPE2", $alignmentScore);
    }
    #else SPE tests failed

    #check if the two sentances are similar
    $alignmentScore = &similarityRule(\@contentWordsA, \@contentWordsB, 
				      \@contentPOSTagsA, \@contentPOSTagsB,
				      \@unmatchedContentWordsA, \@unmatchedContentWordsB,
				      \@unmatchedContentPOSTagsA, \@unmatchedContentPOSTagsB,
				      \@conditions, $chunkSimilarity);
    if ($alignmentScore > 0) {
	return ("SIMI", $alignmentScore);
    }

    #lower limits on chunk similarity for relatedness scores of 4 and 3
    my $relatedThreshold4 = 0.5;
    my $relatedThreshold3 = 0.4;

    #all other tests failed, the sentances are just related in some unknown manner
    if ($chunkSimilarity > $relatedThreshold4)  { 
	return ("REL", 4);
    }
    elsif ($chunkSimilarity > $relatedThreshold3) {
	return ("REL", 3);
    }
    else {
	return ("REL", 2);  #why not (REL, 1)?  (...paper says REL 2)
    }
}


############################
#---------- Rules ----------
############################

#Applies conditions (These are used for multiple subsequent rules) based on the NeroSim paper, used in subsequent 
#      alignment reasoning tests
#   Input - \@chunkA, \@chunkB, \@posTagsA, \@posTagsB, \@contentWordsA, 
#             \@contentWordsB, \@contentPOSTagsA, \@contentPOSTagsB
#          = references to two chunks, their part of speech (POS) tags, the content words of the chunks,
#                   the POS tags of the content words
#   Returns - an array of true/false values corresponding to the results of the conditions
sub applyConditions
{
    #grab variables
    my @chunkA = @{ shift @_ };
    my @chunkB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my @contentWordsA = @{ shift @_ };
    my @contentWordsB = @ { shift @_ };
    my @contentPOSTagsA = @{ shift @_  };
    my @contentPOSTagsB = @{ shift @_ };
    my @conditions = ();

    #apply conditions explained in the NeroSim paper
    #condition 0 - A has a conjunction, B does not
    $conditions[0] = (&hasConjunction(\@chunkA, \@posTagsA)
		      ne &hasConjunction(\@chunkB, \@posTagsB));

    #condition 1 - A and B have content words that are antonyms
    $conditions[1] = &containAntonymContentWords(\@contentWordsA, \@contentWordsB,
						 \@contentPOSTagsA, \@contentPOSTagsB);

    #condition 2 - A or B contain numeric entities
    $conditions[2] = &hasNumericQuantity(\@chunkA, \@posTagsA) 
	|| &hasNumericQuantity(\@chunkB, \@posTagsB);

    #condition 3 - A and B have locations
    $conditions[3] = &hasLocation(\@chunkA, \@posTagsA) 
	&& &hasLocation(\@chunkB, \@posTagsB);

    #condition 4 - A or B have dates or times
    $conditions[4] = &hasDateTime(\@chunkA, \@posTagsA) && &hasDateTime(\@chunkB, \@posTagsB);

    #condition 5 - A and B share non-noun content words
    $conditions[5] = &shareNonNounContentWords(\@contentWordsA, \@contentWordsB, 
					       \@contentPOSTagsA, \@contentPOSTagsB);

    #condition 6 - either chunk has a conjunction
    $conditions[6] = &hasConjunction(\@chunkA, \@posTagsA) 
	|| &hasConjunction(\@chunkB, \@posTagsB);
    
    #return the conditions
    return @conditions;
}


#Applies a sentance equivalency rule based on the NeroSim paper
#   Input - $aContentCount, $bContentCount = the number of content words in chunks A and B
#           - \@unmatchedPOSTagsA, \@unmatchedPOSTagsB - lists of part of speech (POS) tags for the unmatched
#                    content words in chunks A and B
#           - \@conditions - an array of boolean values for each condition (per NeroSim) paper
#           - $similarityScore - the pre-computed similarity score between the two chunks
#   Returns - 1 if the two sentance are equivalent, 0 otherwise 
sub equivalentRule
{
    #Grab Input
    my $aContentCount = shift;
    my $bContentCount = shift;
    my @unmatchedPOSTagsA = @{ shift @_ };
    my @unmatchedPOSTagsB = @{ shift @_ };
    my @conditions = @ {shift @_ };
    my $chunkSimilarity = shift;

    #check if all words have a match
    if (scalar @unmatchedPOSTagsA == 0 && scalar @unmatchedPOSTagsB == 0) {
	return 1;  #all content words in the chunks matched, return true
    }

    #Check precursor conditions for further testing
    if ($conditions[0] && $conditions[1] && $conditions[2] && $conditions[3] && $conditions[4]) {
	my $equivalentThreshold = 0.6;

	#check if the number of content words are the same and the assigned similarity score high
	if ($chunkSimilarity > $equivalentThreshold && $aContentCount == $bContentCount) {
	    return 1;
	}

	#check that if only a single chunk has an unmatched word that it is a proper noun
	if (scalar @unmatchedPOSTagsA == 0 || scalar @unmatchedPOSTagsB == 0) {
	    foreach my $posTag(@unmatchedPOSTagsA) {
		if (not &isProperNoun($posTag)) {
		    return 0;
		}
	    }
	    foreach my $posTag(@unmatchedPOSTagsB) {
		if (not &isProperNoun($posTag)) {
		    return 0;
		}
	    }

	    #the unmatched content words were proper nouns, return true;
	    return 1;
	}
    }

    #all tests failed, return false
    return 0;

}

#Applies a sentance opposite rule based on the NeroSim paper
#   Input - \@conditions = boolean values for the conditions applied (per NeroSim paper)
#   Returns - 1 if the chunks are opposites, 0 otherwise
sub oppositeRule
{
    #Grab Input
    my @conditions = @_;

    #only apply opposite rules if these conditions are false
    #    ...so if there are no numbers and no conjunctions
    if ($conditions[2] || $conditions[6]) {
	return 0;
    }
    
    #they are opposite only if one chunk contains an antonym of the other chunk
    return $conditions[1];
}


#Determines one sentance has the same general content but is more specific than the other.
#   Also determines the directions of that relationship
#   Input - \@chunkA, \@chunkB, \@posTagsA, \@posTagsB, \contentA, \@contentB, \@contentPOSA, \@contentPOSB
#              = tokenized chunks, their corresponding part of speech (POS) Tags, contentWords, and POS Tags of the content words
#   Returns - ($specificity direction, $score) - the direction of the specificity (A more than B or B more than A) and the score
#                        of the specificty.  (Direction is based on constant in class). 
#               - returns (0, 0) if the specificity rule fails
sub specificityRule
{
    #Grab Input
    my @chunkA = @{ shift @_ };
    my @chunkB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my @contentA = @{ shift @_ };
    my @contentB = @ { shift @_ };
    my @contentPOSA = @ { shift @_ };
    my @contentPOSB = @ {shift @_ };


    #Determine if A is more specific than B and vice-versa based on content word counts
    my $score =&specificityRuleHelper(\@chunkB, \@chunkA, \@posTagsB, \@posTagsA, 
				 \@contentB, \@contentA, \@contentPOSB, \@contentPOSA);
    if ($score) {
	return ($B_More_Specific_Than_A, $score);
    }

    $score = &specificityRuleHelper(\@chunkA, \@chunkB, \@posTagsA, \@posTagsB, 
				      \@contentA, \@contentB, \@contentPOSA, \@contentPOSB);
    if ($score) {
	return ($A_More_Specific_Than_B, $score);
    }

    #Determine which is more specific based on hyper/hyponymy
    #determine the number of nouns in A and B
    my $aCount = scalar @contentA;
    my $aNounCount = 0;
    my $aNoun;
    my $aNounPOSTag;
    for (my $i = 0; $i < $aCount; $i++) {
	if (&isNoun($contentPOSA[$i])) {
	    $aNounCount++;
	    $aNoun = $contentA[$i];
	    $aNounPOSTag = $contentPOSA[$i];
	}
    }
    my $bCount = scalar @contentB;
    my $bNounCount = 0;
    my $bNoun;
    my $bNounPOSTag;
    for (my $i = 0; $i < $bCount; $i++) {
	if (&isNoun($contentPOSB[$i])) {
	    $bNounCount++;
	    $bNoun = $contentB[$i];
	    $bNounPOSTag = $contentPOSB[$i];
	}
    }

    #if each sentance contains a single noun see if one is a hyper/hyponym of the other
    if ($aNounCount == 1 && $bNounCount == 1) {
	if (&isHypernym($aNoun, $bNoun, $aNounPOSTag, $bNounPOSTag)) {
	    return ($B_More_Specific_Than_A, 3);
	}
	if (&isHypernym($bNoun, $aNoun, $bNounPOSTag, $aNounPOSTag)) {
	    return ($A_More_Specific_Than_B, 3);
	}
    }
	
    #all tests failed, return false
    return (0, 0);
}


#Compares two chunks A and B and determines if one is more specific than the other based on them containing the same content words and then extra words
#    Input - \@chunkA, \@chunkB, \@posTagsA, \@posTagsB, \contentA, \@contentB, \@contentPOSA, \@contentPOSB
#              = tokenized chunks, their corresponding part of speech (POS) Tags, contentWords, and POS Tags of the content words
#   Returns - the specificity score (per SEMEVAL requirements) or 0 if specifictiy test fails
sub specificityRuleHelper
{
    #grab input
    my @chunkA = @{ shift @_ };
    my @chunkB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @ { shift @_ };
    my @contentA = @{ shift @_ };
    my @contentB = @{ shift @_ };
    my @contentPOSA = @{ shift @_ };
    my @contentPOSB = @{ shift @_ };
    my $aCount = scalar @contentA;
    my $bCount = scalar @contentB;

    #determine if A contains all the content words of B
    my $aHasAllB = 1;
    #find each content word of B in A
    for (my $i = 0; $i < $bCount; $i++) {
	my $match = 0;
	for (my $j = 0; $j < $aCount; $j++) {
	    if (&wordsEquivalent($contentA[$j], $contentB[$i], $contentPOSA[$j], $contentPOSB[$i])) {
		$match = 1;
		last;
	    }
	}
	#tokenB is not in A, therefore A does not contain all the contents words of B, done
	if (not $match) {
	    $aHasAllB = 0;
	    last;
	}  
    }

    #tests passed, A contains all the content words of B
    if ($aHasAllB) {
	#check if one is more specific based on the presence of a conjunction
	my $aHasConj = &hasConjunction(\@chunkA, \@posTagsA);
	my $bHasConj = &hasConjunction(\@chunkB, \@posTagsB);
	if ($aHasConj && not $bHasConj) {
	    return 4;
	} 

	#determine if A has any non-verb content words (implying adjective, adverb, or extra noun)
	foreach my $posTag(@contentPOSB) {
	    if (not &isVerb($posTag)) {
		return 3;
	    }  
	}
    }
    
  
    #tests failed return 0
    return 0;
}


#Applies a Rule from the NeroSim paper to determine if two chunks are similar
#   Input - \@contentWordsA, \@contentWordsB, \@contentPOSTagsA, \@contentPOSTagsB
#             \@unmatchedContentWordsA, \@unmatchedContentWordsB, \@unmatchedContentPOSTagsA, 
#             \@unmatchedContentPOSTagsB, \@conditions, $similarityScore
#           = the content words, corresponding part of speech (POS) tags, the unmatched content words, their 
#                   corresponding POS tags of two chunks (A and B).  A list of boolean condition values (per NeroSim paper),
#                   and a per-computed similarity score between the two chunks
#   Returns - the alignment score (per SEMEVAL requirements), or 0 if the similairty rule fails
sub similarityRule
{
    my @contentWordsA = @ { shift @_ };
    my @contentWordsB = @ { shift @_ };
    my @contentPOSTagsA = @ { shift @_ };
    my @contentPOSTagsB = @ { shift @_ };
    my @unmatchedContentWordsA = @ { shift @_ };
    my @unmatchedContentWordsB = @ { shift @_ };
    my @unmatchedContentPOSTagsA = @ { shift @_ };
    my @unmatchedContentPOSTagsB  = @ { shift @_ };		       	   
    my @conditions = @ { shift @_ };
    my $chunkSimilarity =  shift;

    #check if the only difference is a number
    my $onlyNumbersDiffer = 1;
    #check unmatched content word in A
    foreach my $token(@unmatchedContentWordsA) {  
	if (not $wq->isNumber($token)) {
	    $onlyNumbersDiffer = 0;
	    last;
	}
    }
    if ($onlyNumbersDiffer) {
	#check each unmatched content word in B
	foreach my $token(@unmatchedContentWordsB) {
	    if (not $wq->isNumber($token)) {
		$onlyNumbersDiffer = 0;
		last;
	    }
	}
	#all content words checked, if all words have been numbers return
	if($onlyNumbersDiffer) {
	    return 3;
	}	
    }

    #check if both contain a date or time or both contain a location
    if ($conditions[4] || $conditions[3]) {
	return 3;
    }

    #lower limit chunk similarities for assigning a similarity
    #   score of 4 and 3  
    my $similarityThreshold4 = 0.5;
    my $similarityThreshold3 = 0.4;


    #check if both chunks share a noun
    my $numContentWordsA = scalar @contentWordsA;
    my $numContentWordsB = scalar @contentWordsB;
    #check each word in A against each word in B
    for (my $i = 0; $i < $numContentWordsA; $i++) {
	for (my $j = 0; $j < $numContentWordsB; $j++) {
	    #only compare nouns
	    if (&isNoun($contentPOSTagsA[$i]) && &isNoun($contentPOSTagsB[$j])) {
		#test equivalency
		if (&wordsEquivalent($contentWordsA[$i], $contentWordsB[$j], 
				     $contentPOSTagsA[$i], $contentPOSTagsB[$j])) {
		    #words match, return a score based on the chunk similarity
		    if ($chunkSimilarity >= $similarityThreshold3) {  
			return 3;
		    }
		    else {
			return 2;
		    }
		}
	    }
	}
    }

    #check non-noun content word sharing (condition 5)
    if ($conditions[5]) {
	#...the chunks contain non-noun content words, 
	#    return a similarity score based on the chunk similarity
	if ($chunkSimilarity > $similarityThreshold4) {
	    return 4;
	}
	if ($chunkSimilarity > $similarityThreshold3) {
	    return 3;
	}
	return 2;
    }
    
    #all tests failed, return false
    return 0;
}



#########################################
#---------- Chunk Comparison Methods ----------
#########################################

#Gets a list of content words that do not have a matching content word in the other chunk
#   two words match if they are equivalent, direct or indiret synonyms
#   Input - \@contentA, \@contentB, \@posTagsA, \@posTagsB = the content words of two chunks and the part of speech (POS) tags for the content words
#   Returns - (\@unmatchedTokensA, \@unmatchedTokensB, \@unmatchedPOSTagsA, \@unmatchedPOSTagsB)
#               =  a list of unmatched content words in chunkA, a list of unmatched content words in chunkB, and their corresponding POS tags
sub getUnmatchedContentWords
{
    #grab input
    my @contentA = @{ shift @_ };
    my @contentB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my $aCount = scalar @contentA;  
    my $bCount = scalar @contentB;

    #create arrays to keep track of if tokens are matched
    my @aMatched = ();
    my @bMatched = ();
    #initialize all to unmatched
    foreach my $token(@contentA) {
	push @aMatched, 0;
    }
    foreach my $token(@contentB) {
	push @bMatched, 0;
    }

    #Find matching content words...this is split into two loops to improve run time 
    #    (generally there will only be a sinlge match (if any) between content words)
    #check each token in A against each token in B and try to find any matches
    for (my $i = 0; $i < $aCount; $i++) {
	my $tokenA = $contentA[$i];
	my $posA = $posTagsA[$i];

	#check this token against each token in B
	for (my $j = 0; $j < $bCount; $j++) {
	    my $tokenB = $contentB[$j];
	    my $posB = $posTagsB[$j];

	    if ($wq->isSynonym(&toWordNetFormat($tokenA, $posA), &toWordNetFormat($tokenB, $posB))) {
		$bMatched[$j]++;  #record that b matched
		$aMatched[$i]++;  #record that a matched
		last;  #your done with this word
	    }
	}	
    }

    #check any remaining tokens in B against tokens in A to find matches
    for (my $i = 0; $i < $bCount; $i++) {
	my $tokenB = $contentB[$i];
	my $posB = $posTagsB[$i];

	if ($bMatched[$i] < 1) {
	    #check this token against each token in A
	    for (my $j = 0; $j < $aCount; $j++) {
		my $tokenA = $contentA[$j];
		my $posA = $posTagsA[$j];

		if ($wq->isSynonym(&toWordNetFormat($tokenA, $posA), &toWordNetFormat($tokenB, $posB))) {
		    $aMatched[$i]++;  #record that a matched
		    $bMatched[$j]++;  #record that b matched
		    last;  #your done with this word
		}
	    }
	}	
    }

    #build a list of unmatched A tokens and POS tags
    my @unmatchedTokensA = ();
    my @unmatchedPOSTagsA = ();
    for (my $i = 0; $i < $aCount; $i++) {
	if ($aMatched[$i] < 1) {
	    push (@unmatchedTokensA, $contentA[$i]);
	    push (@unmatchedPOSTagsA, $posTagsA[$i]);
	}
    }
    #build a list of unmatched B tokens and POS tags
    my @unmatchedTokensB = ();
    my @unmatchedPOSTagsB = ();
    for (my $i = 0; $i < $bCount; $i++) {
	if ($bMatched[$i] < 1) {
	    push (@unmatchedTokensB, $contentB[$i]);
	    push (@unmatchedPOSTagsB, $posTagsB[$i]);
	}
    }

    #return the results
    return (\@unmatchedTokensA, \@unmatchedTokensB,\@unmatchedPOSTagsA, \@unmatchedPOSTagsB);
}


#Determines if one chunk contains content words that are antonyms of the content words in another chunk
#   Input -  \@contentA, \@contentB, \@ posTagsA, \@posTagsB 
#           = references to 4 arrays containing two lists of content words and the corresponding part of speech (POS) tags.  
#   Returns - 1 if the chunks have antonym contents words, 0 otherwise
sub containAntonymContentWords
{
    #Grab the input
    my @contentA = @{shift @_ };
    my @contentB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my $numAWords = scalar @contentA;
    my $numBWords = scalar @contentB;

    #compare each content word and
    for (my $i = 0; $i < $numAWords; $i++) {
	for (my $j = 0; $j < $numBWords; $j++) {

	    #see if the words are antonyms 
	    #Note: sine we are only looking at content words they will be of wordnet type
	    if ($wq->isAntonym(&toWordNetFormat($contentA[$i], $posTagsA[$i])
		     , &toWordNetFormat($contentB[$j], $posTagsB[$j])) > 0) {
		return 1;  #the words are antonyms so return true
	    }
	}
    }

    #no words were antonyms, return false
    return 0;
}


#Determines if two chunks share any non-content words
#   Input -  \@contentA, \@contentB, \@ posTagsA, \@posTagsB 
#           = references to 4 arrays containing two lists of content words and the corresponding part of speech (POS) tags.  
#   Returns - 1 if the chunks share any non-noun contents words, 0 otherwise
sub shareNonNounContentWords
{
    #Grab the input
    my @contentA =@{ shift @_ };
    my @contentB = @{ shift @_ };
    my @posTagsA = @{ shift @_ };
    my @posTagsB = @{ shift @_ };
    my $numAWords = scalar @contentA;
    my $numBWords = scalar @contentB;

    #compare the content words
    for (my $i = 0; $i < $numAWords; $i++) {
	for (my $j = 0; $j < $numBWords; $j++) {
	    #only compare non-nouns
	    if (not (&isNoun($posTagsA[$i])) && not (&isNoun($posTagsB[$j]))) {
		#compare the words
		if (&wordsEquivalent($contentA[$i], $contentB[$j], $posTagsA[$i], $posTagsB[$j])) {
		    return 1;  #words are equivalent, return true
		}
	    }
	}
    }
   
    #no words matched, return false
    return 0;
}



###########################################
#---------- Single Chunk Helper Functions ----------
###########################################

#Gets just the content words from a chunk
#   Input - a reference to an array of tokens and a reference to a paralell array of POS tags
#   Returns - array references to the content word tokens and the corresponding POS tags        
sub getContentWords
{
    #grab input variables
    my @tokens = @{ shift @_ };
    my @posTags = @{ shift @_ };

    #create return variables
    my @contentWords = ();
    my @contentPOSTags = ();

    #check each token in the chunk
    my $numTokens = scalar @tokens;
    for(my $i = 0; $i < $numTokens; $i++)  {

	#check if the token is a content word
	if (&isContentWord($posTags[$i])) {
	    push (@contentWords, $tokens[$i]);   
	    push(@contentPOSTags, $posTags[$i]);
	}
    }

    return (\@contentWords, \@contentPOSTags);
}


#Determines if the chunk contains a conjunction
#   Input - a reference to an array of part of speech (POS) tags
#   Returns - 1 if the chunk contains a conjunction, 0 otherwise
sub hasConjunction
{   
    #grab the input chunk
    my @posTags = @{ shift @_ };
   
    #check each tag
    foreach my $tag(@posTags) {
	if (&isConjunction($tag)) {
	    return 1; 
	}
    }
    return 0;
}


#Determines if a chunk contains a numeric quantity (not including dates or times)
#   Input - a reference to an array of tokens in a chunk
#   Returns - 1 if the chunk contains a non-dateTime numeric quantity, 0 otherwise
sub hasNumericQuantity
{
    #grab the input chunk
    my @chunk = @{ shift @_ };
    #check each token of the chunk to see if it is a date or a time
    foreach my $token (@chunk) {
	if ($wq->isNumber($token)
	    && not ($wq->isDate($token))
	    && not ($wq->isTime($token))) {
	    return 1;
	}
    }

    #no dates or times found, return false
    return 0;
}


#Determines if a chunk contains a location
#   Input - a reference to an array of tokens in a chunk
#   Returns - 1 if the chunk contains a location, 0 otherwise
sub hasLocation
{
    #grab the input chunk
    my @chunk = @{ shift @_ };

    #check each token of the chunk to see if it is a date or a time
    foreach my $token (@chunk) {
	if ($wq->isLocation($token)) {
	    return 1;
	}
    }

    #no dates or times found, return false
    return 0;
}


#Determines if a chunk contains a date or time token
#  Input - a reference to an array of tokens in a chunk
#  Returns - 1 if the chunk contains a date or a time, 0 otherwise
sub hasDateTime
{
    #grab the input chunk
    my @chunk = @{ shift @_ };

    #check the entire chunk to see if it is a date or time
    #   this applies to phrases such as the fifth of may, etc...
    my $phrase = join(" ", @chunk);
    if ($wq->isTime($phrase) || $wq->isDate($phrase)) {
	return 1;
    }
    
    #check each token of the chunk to see if it is a date or a time
    foreach my $token (@chunk) {
	if ($wq->isTime($token) || $wq->isDate($token)) {
	    return 1;
	}
    }

    #no dates or times found, return false
    return 0;
}



####################################
#---------- Token Helper Methods ----------
####################################

#Determines if two tokens are equivalent meaning the same or very closely related (closer than synonyms)
#   For instance lady and woman are equivalent however woman and girl are not
#   Input - $tokenA, $tokenB, $posTagA, $posTagB - the two tokens to compare and their part of speech (POS) tags
#   Returns - 1 if the tokens are equivalent, 0 otherwise
sub wordsEquivalent
{
    #grab input
    my $tokenA = shift;
    my $tokenB = shift;
    my $posTagA = shift;
    my $posTagB = shift;

    #check if the words have the same POS tag
    if ($posTagA ne $posTagB) {
	return 0;  #POS tags are not the same, words are not the same, return false
    }

    #check if the tokens are identical
    #   this is not only faster than WordNet lookup but is also necassary for puncuation 
    #      and for words not in WordNet (like he, she, the, and, proper names, etc...)
    if ($tokenA eq $tokenB) {
	return 1;  #POS tags and tokens are identical, return true
    }

    #Check if the words are direct synonyms (as in woman and lady)
    #   also we know that the POS tags are the same
    if(&posTagInWordNet($posTagA)) {
	if ($wq->isSynonym(&toWordNetFormat($tokenA, $posTagA), &toWordNetFormat($tokenB, $posTagB), 1)) {
	    return 1;
	}
    }

    #all tests failed, return false
    return 0;
}


#Determines if Token A is a hypernym of Token B
#   Input - $token A, $tokenB, $posTagA, $posTagB = the two words to compare and their corresponding part of speech (POS) tags
#   Returns - 1 if the $tokenA is a hypernym of $tokenB
sub isHypernym
{
    my $tokenA = shift;
    my $tokenB = shift;
    my $posTagA = shift;
    my $posTagB = shift;

    #Check for hypernymy
    #   also we know that the POS tags are the same
    if(&posTagInWordNet($posTagA) && &posTagInWordNet($posTagB)) {
	if ($wq->isHypernym(&toWordNetFormat($tokenA, $posTagA), &toWordNetFormat($tokenB, $posTagB))) {
	    return 1;
	}
    }   

    return 0;
}


#Determines if a token is a content word based on its POS tag
#   Input - a part of speech tag for a token
#   Returns - 1 if it is a content word, 0 otherwise
sub isContentWord
{
    #grab input
    my  $posTag = shift;

    #determine if the token is a content word
    if (&isNoun($posTag)
	|| &isVerb($posTag)
        || &isAdjective($posTag)
	|| &isAdverb($posTag)) {   #other POS tags?   ...need to be sure which pos tagger we are using
	return 1;
    }
    return 0;
}


#Determines if a token is a proper noun
#   Input - a part of speech tag for a token
#   Returns - 1 if the token is a proper noun, 0 otherwise
sub isProperNoun 
{
    my $token = shift;
    if ($token =~ m/$POS_ProperNoun/) {
	return 1;
    }
    return 0;
}

#Determines if a token is a noun
#   Input - a part of speech tag for a token
#   Returns - 1 if the token is a noun, 0 otherwise
sub isNoun
{
    my $token = shift;
    if ($token =~ m/$POS_Noun/) {
	return 1;
    }
    return 0;

}

#Determines if a token is a noun
#   Input - a part of speech tag for a token
#   Returns - 1 if the token is a verb, 0 otherwise
sub isVerb
{
    my $token = shift;
    if ($token =~ m/$POS_Verb/) {
	return 1;
    }
    return 0;
}

#Determines if a token is an adjective
#   Input - a part of speech tag for a token
#   Returns - 1 if the token is an adjective, 0 otherwise
sub isAdjective
{
    my $token = shift;
    if ( $token =~ m/$POS_Adjective/) {
	return 1;
    }
    return 0;
}

#Determines if a token is an adjective
#   Input - a part of speech tag for a token
#   Returns - 1 if the token is an adjective, 0 otherwise
sub isAdverb
{
    my $token = shift;
    if ( $token =~ m/$POS_Adverb/) {
	return 1;
    }
    return 0;
}

#Determines if a token is an conjunction
#   Input - a part of speech tag for a token
#   Returns - 1 if the token is a conjunction, 0 otherwise
sub isConjunction
{
    my $token = shift;
    if ($token =~ m/$POS_Conjunction/) {
	return 1;
    }
    return 0;
}


#Checks if the part of speech (POS) type (based on a tag) is in WordNet
#   Input - a pos tag
#   Returns - 1 if the pos tag is in WordNet, 0 otherwise
sub posTagInWordNet
{
    my $posTag = shift;

    #only nouns, verbs, adjectives, and adverbs are in wordnet
    if (&isNoun($posTag)
	|| &isVerb($posTag)
	|| &isAdjective($posTag)
	|| &isAdverb($posTag)) {
	return 1;
    }
    else {
	return 0;
    }
}


#Makes a token into WordNet format (word#pos). If the POS tag is not in WordNet just the token is returned
#   Input a token and its part of speech (POS)
#   Returns a string in WordNet format for the token and POS
sub toWordNetFormat
{
    #grab the input
    my $token = shift;
    my $posTag = shift;

    if ($token !~ /#[n|v|a|r]/) { #check if POS tag is already appended
	#check each POS tag in WordNet and append
	if (&isNoun($posTag)) {
	    return $token."#n";
	}
	if (&isVerb($posTag)) {
	    return $token."#v";
	}
	if (&isAdjective($posTag)) {
	    return $token."#a";
	}
	if (&isAdverb($posTag)) {
	    return $token."#r";
	}
	else {
	    return $token;  #unknown, don't append POS tag
	}
    }

    #token is already in the correct format (with POS tag)
    return $token;
}




=comment
    @Author = Sam Henry
    @Description = performs rule based classificaiton of pre-aligned chunks to assign alignment reasoning and alignment scores.  The idea is based on the NeroSim paper
=cut
