#use First Part
use ChunkAlignment;
use AlignmentReasoning;
use OutputResults;
use WordQuery;
use ModulePreprocess;
use ReadGoldOutput;

use WordNet::Similarity::vector;
use Lingua::EN::Tagger;
use WordNet::QueryData;
use WordNet::SenseRelate::AllWords;
use List::Util qw( min max );

use strict;
use warnings;

#---------------------------------------------------------------------------------------------------
#    Hard Coded Parameters
#--------------------------------------------------------------------------------------------------

#Chunk Alignment Parameters
#if two words have a similarity score lower than this then their score is set to 0
my $wordToWordSimilarityThreshold = 0.9;
#if two chunks have a similarity score lower than this then their score is set to 0
my $chunkSimilarityThreshold = 0.0;
my $maxLevenshtein = 0;  #the max levenshtein edit distance before setting to 0.  if set to 0 only exact word matches will be used
my $useWSD = 0;  #toggles WSD for chunk alignment
my $useStopWordRemoval = 0;
my $rechunk = 0;

# headlines dataset files
my $headlinesFileName1 = 'data/STSint.input.headlines.sent1.chunk.txt';
my $headlinesFileName2  = 'data/STSint.input.headlines.sent2.chunk.txt';
my $headlinesGoldStandardFileName = 'data/STSint.input.headlines.wa';
my $headlinesResultsFileName = 'results/headlinesTrainingOut.txt';
my $headlinesSamplesOutFilePath = 'results/headlinesSamples.csv';

my $headlinesTestFileName1 = 'data/test/STSint.testinput.headlines.sent1.chunk.txt';
my $headlinesTestFileName2 = 'data/test/STSint.testinput.headlines.sent2.chunk.txt';
my $headlinesTestOutput = 'results/headlinesTestOut.txt';

#Images dataset files
my $imagesFileName1 = 'data/STSint.input.images.sent1.chunk.txt';
my $imagesFileName2 = 'data/STSint.input.images.sent2.chunk.txt';
my $imagesGoldStandardFileName = 'data/STSint.input.images.wa';
my $imagesResultsFileName = 'results/rerun/imagesTrainingOut.txt';
my $imagesSamplesOutFilePath = 'results/rerun/imageSamplesOut.csv';

my $imagesTestFileName1 = 'data/test/STSint.testinput.images.sent1.chunk.txt';
my $imagesTestFileName2 = 'data/test/STSint.testinput.images.sent2.chunk.txt';
my $imagesTestOutput = 'results/rerun/imagesTestOut.txt';

#Answers dataset files
my $answersFileName1 = 'data/STSint.input.answers-students.sent1.chunk.txt';
my $answersFileName2 = 'data/STSint.input.answers-students.sent2.chunk.txt';
my $answersGoldStandardFileName = 'data/STSint.input.answers-students.wa';
my $answersResultsFileName = 'results/answersTrainingOut.txt';
my $answersSamplesOutFilePath = 'results/answersSamplesOut.csv';

my $answersTestFileName1 = 'data/test/STSint.testinput.answers-students.sent1.chunk.txt';
my $answersTestFileName2 = 'data/test/STSint.testinput.answers-students.sent2.chunk.txt';
my $answersTestOutput = 'results/answersTestOut.txt';


#----------------------------------------------------------------------
#      Program Code
#----------------------------------------------------------------------

#initialize variables
my $qd = WordNet::QueryData->new;

#Initialize WSD
my $wntools = WordNet::Tools->new($qd);
my %options = (wordnet => $qd, wntools => $wntools, measure => 'WordNet::Similarity::vector');
my $wsd = WordNet::SenseRelate::AllWords->new (%options);  	

#initialize POS Tagger
my $tagger = new Lingua::EN::Tagger;
#initialize Similarity Measure used in chunk alignment
my $measure=WordNet::Similarity::vector->new($qd);
my $wq = WordQuery->new(); #initialize WordQuery


#perform comparisons
&compareFiles($headlinesFileName1, $headlinesFileName2, $headlinesResultsFileName);
&compareFiles($headlinesTestFileName1, $headlinesTestFileName2, $headlinesTestOutput);

&compareFiles($imagesFileName1, $imagesFileName2, $imagesResultsFileName);
&compareFiles($imagesTestFileName1, $imagesTestFileName2, $imagesTestOutput);

&compareFiles($answersFileName1, $answersFileName2, $answersResultsFileName);
&compareFiles($answersTestFileName1, $answersTestFileName2, $answersTestOutput);

#Use this code block to convert each chunk pair to a set of features for use in WEKA
#convert the gold standard chunks to features for use in WEKA
#&convertGoldToFeatures($headlinesGoldStandardFileName, $headlinesSamplesOutFilePath);
#&convertGoldToFeatures($imagesGoldStandardFileName, $imagesSamplesOutFilePath);
#&convertGoldToFeatures($answersGoldStandardFileName, $answersSamplesOutFilePath);


#--------------------------------------------------------------------
#      Begin Subroutines
#--------------------------------------------------------------------

#Compares two files, performs chunk alignment and applies rule-based alignment reasoning
sub compareFiles
{
    #grab variables
    my $file1 = shift;
    my $file2 = shift;
    my $resultsFileName = shift;
    my $ruleSetIndex = shift;
    my $scoreTypeIndex = shift;

    #open file
    open FILE1, $file1; 
    open FILE2,  $file2;

    #loop through all the sentences in the two files
    my $sentenceID = 0;
    while ( my $sentence1 = <FILE1>) {  #note:  this loop assumes the two input files are the same length

	#update the sentence ID (used as a key for determinining accuracy)
	$sentenceID++;
	
	#grab sentence 1 (done at the top of the while loop)
	chomp $sentence1;
	my @chunkArray1 = &parseSentence($sentence1);

	#grab sentence 2
	my $sentence2 = <FILE2>;
	chomp $sentence2;
	my @chunkArray2 = &parseSentence($sentence2);

	#ChunkAlignment requires that sentence1 is longer than or equal length of sentence2
	my $swapped = 0;
	my $sentence1Length = scalar @chunkArray1;
	my $sentence2Length = scalar @chunkArray2;
	if ($sentence1Length < $sentence2Length) {
	    my @temp = @chunkArray1;
	    @chunkArray1 = @chunkArray2;
	    @chunkArray2 = @temp;
	    $swapped = 1;
	}
	#----------------
	#align the chunks of the two sentences
	my @results = 
	    &ChunkAlignment::alignChunks(\@chunkArray1, \@chunkArray2, 
					 $qd, $wntools, $wsd, $measure, $tagger, $wq,
					 $wordToWordSimilarityThreshold, $chunkSimilarityThreshold,
					 $maxLevenshtein, $useWSD, $useStopWordRemoval, $rechunk);

	#grab the output needed for output
	my @words1 = @{ $results[0] };
	my @words2 = @{ $results[1] };
	my @chunks1 = @ { $results[2] };
	my @chunks2 = @ { $results[3] };
	my @posTags1 = @{ $results[4] };
	my @posTags2 = @{ $results[5] };
	my @alignments = @ { $results[6] };
	my @chunkSimilarities = @ { $results[7] };

	#-----------
	#determine the alignment reasoning
	my @alignmentReasons = 
	    &AlignmentReasoning::JRip_Rules(\@words1, \@words2,
					    \@chunks1, \@chunks2,
					    \@posTags1, \@posTags2, 
					    \@alignments, \@chunkSimilarities,
					    $wq, $ruleSetIndex);

	#if the sentences are swapped we need to swap SPE1 and SPE2 alignment reasons
	if ($swapped > 0) {
	    for (my $i = 0; $i < scalar @alignmentReasons; $i++) {
		if ($alignmentReasons[$i] eq 'SPE1') {
		    $alignmentReasons[$i] = 'SPE2';
		} 
		elsif ($alignmentReasons[$i] eq 'SPE2') {
		    $alignmentReasons[$i] = 'SPE1';
		}
	    }
	}

	#---------
	#assign the alignemnt score
	#these numbers are the averages of the images dataset
	#   These were found to produce the best results for all
	#   datasets
	my $equiAverage = 5;
	my $oppoAverage = 4.00;
	my $spe1Average = 3.74; 
	my $spe2Average = 3.69;
	my $simiAverage = 2.975;
	my $relAverage = 3;

	my @alignmentScores = ();
	foreach my $alignmentReason(@alignmentReasons) {
	    if ($alignmentReason eq 'EQUI') {
		push @alignmentScores, $equiAverage;
	    }
	    elsif ($alignmentReason eq 'OPPO') {
		push @alignmentScores, $oppoAverage;
	    }
	     elsif ($alignmentReason eq 'SPE1') {
		push @alignmentScores, $spe1Average;
	    }
	     elsif ($alignmentReason eq 'SPE2') {
		push @alignmentScores, $spe2Average;
	    }
	     elsif ($alignmentReason eq 'SIMI') {
		push @alignmentScores, $simiAverage;
	    }
	     elsif ($alignmentReason eq 'REL') {
		push @alignmentScores, $relAverage;
	    }
	    else {
		push @alignmentScores, 0;
	    }
	}

	#---------------
	#output the results
	&outputResults(\@words1, \@words2, \@chunks1, \@chunks2, 
		       \@alignments, \@alignmentReasons, \@alignmentScores, 
		       $sentenceID, $resultsFileName, $swapped);
    }
}


#reads in a gold standards file, converts each chunk pair to a set of features
#   and outputs the features for use in an external program (WEKA)
sub convertGoldToFeatures
{
    #grab variables
    my $goldStandardFileName = shift;
    my $samplesOutFilePath = shift;

    #grab the gold standard for this pair
    my @solutions = @{ &readGold( $goldStandardFileName) };
    my $numSolutions = scalar @solutions;

    #extract the features for each chunk pairs provided
    my $sentenceID = 0;
    my @allSamples = ();
    my @trueClasses = ();
    my @sentenceIDs = ();
    my @trueScores = ();
    my @swapped = ();
    for ($sentenceID = 0; $sentenceID < $numSolutions; $sentenceID++) {
	print "converting $sentenceID to features\n";

	#grab the solution data structures
	my %solution = %{ $solutions[$sentenceID] };
	my @words1 = @{ $solution{words1} };
	my @words2 = @{ $solution{words2} };
	my @chunks1 = @{ $solution{chunks1} };
	my @chunks2 = @{ $solution{chunks2} };
	my @alignments = @{ $solution{alignments} };       
	my @alignmentReasons = @{ $solution{alignmentReasons} };   
	my @alignmentScores = @{ $solution{alignmentScores} };   
	my $swapped = $solution{swapped};
	my $fact = $solution{fact};   
	my $pol = $solution{polarity};   


	#------------- Feature Extraction --------------------------
	#extract other features from the sentences
	my @posTags1 = &ChunkAlignment::getPOSTags(join(" ", @words1), $tagger);
	my @posTags2 = &ChunkAlignment::getPOSTags(join(" ", @words2), $tagger);

	#standardize the tokens
	@words1 = &ChunkAlignment::standardizeTokens(\@words1);
	@words2 = &ChunkAlignment::standardizeTokens(\@words2);

	#get disambiguated words (for use with chunk alignment score)
	my @wordNetTerms1;
	my @wordNetTerms2;
	if ($useWSD) {
	    @wordNetTerms1 = &ChunkAlignment::disambiguateTokens(\@words1, \@posTags1, $wsd);
	    @wordNetTerms2 = &ChunkAlignment::disambiguateTokens(\@words2, \@posTags2, $wsd);
	} else {
	    @wordNetTerms1 = @words1;
	    @wordNetTerms2 = @words2;
	}

	#remove stop words
	if ($useStopWordRemoval) {
	    @words1 = &ChunkAlignment::removeStopWords(\@words1, \@posTags1);
	    @words2 = &ChunkAlignment::removeStopWords(\@words2, \@posTags2);
	}


	#Set thresholds to zero so that the chunk similarities can be used in rules
	$wordToWordSimilarityThreshold = 0;
	$chunkSimilarityThreshold = 0;

	#compute word to word similarities
	my @similarities = @{ &ChunkAlignment::computeWordToWordSimilarities(\@words1, \@words2,
									     \@wordNetTerms1, \@wordNetTerms2,
									     $measure, $wq,
									     $wordToWordSimilarityThreshold, 
									     $maxLevenshtein) };

	#compute chunk to chunk similarities
	my @chunkSimilarities = ();
	for (my $i = 0; $i < scalar @chunks1; $i++) {

	    #get the current chunk Number
	    my $chunk1Number = $chunks1[$i];
	    my $chunk2Number = $alignments[$chunks1[$i]]; 

	    #get the similarity score for this chunk
	    push(@chunkSimilarities, &ChunkAlignment::computeChunkSimilarity(\@chunks1, \@chunks2, 
									     \@similarities, 
									     $chunk1Number, $chunk2Number,
									     $chunkSimilarityThreshold));
	}

	#convert each chunk to a feature list
	my @samples = @{ &AlignmentReasoning::convertToFeatures(\@words1, \@words2,
								\@chunks1, \@chunks2, 
								\@posTags1, \@posTags2, 
								\@alignments, \@alignmentScores,
								\@alignmentReasons, \@chunkSimilarities, 
								$wq) };

      	#if the sentences are swapped we need to swap SPE1 and SPE2 alignment reasons
	if ($swapped > 0) {
	    for (my $i = 0; $i < scalar @alignmentReasons; $i++) {
		if ($alignmentReasons[$i] eq 'SPE1') {
		    $alignmentReasons[$i] = 'SPE2';
		} 
		elsif ($alignmentReasons[$i] eq 'SPE2') {
		    $alignmentReasons[$i] = 'SPE1';
		}
	    }
	}

	#save this sample
	for (my $i = 0; $i < scalar @samples; $i++) {
	    push(@allSamples, $samples[$i]);
	    push(@sentenceIDs, $sentenceID);
	    push(@trueClasses, $alignmentReasons[$i+1]);
	    push(@trueScores, $alignmentScores[$i+1]);
	    push(@swapped, $swapped);
	}
    }

    #------------------- Sample Output -----------------------------
    my @tempSample = @{ $allSamples[0] };
    my $numFeatures = scalar @tempSample;

    #Output the Samples
    open (OUT, ">".$samplesOutFilePath);
    #print column labels
    print OUT "class,true-score,sentenceID,";
    for (my $i = 0; $i < ($numFeatures-1); $i++) {
	print OUT "feature $i,"
    }
    my $lastFeatureNum = $numFeatures-1;
    print OUT "feature $lastFeatureNum\n";
    
    #print out each sample
    for (my $i = 0; $i < scalar @allSamples; $i++) {
	if ($trueClasses[$i] ne 'NOALI') {
	    my $sentenceIDPluseOne = $sentenceIDs[$i] + 1;
	    print OUT "$trueClasses[$i],";
	    print OUT "$trueScores[$i],";
	    print OUT "$sentenceIDs[$i],";

	    #print out the features
	    my @sample = @{ $allSamples[$i] };
	    print OUT join(',', @sample);
	    print OUT "\n";
	}
    }	        
}


=comment
@author = Sam Henry
entry point for the program.  Reads input, aligns chunks, assigns chunk reasoning, and outputs the results
=cut
